#ifndef MpsReadout_HH
#define MpsReadout_HH

// TO DO; Add broadcasts, turn off all reads with cfg data enable



#include <iostream>
#include <fstream>

#include "RcdUserRW.hh"

#include "SubRecordType.hh"
#include "SubInserter.hh"
#include "SubAccessor.hh"

#include "UsbDaqDevice.hh"


class MpsReadout : public RcdUserRW {

public:
  MpsReadout(unsigned s=0, unsigned n=1) {
    for(unsigned i(0);i<n;i++) {
      _vUsbDaq.push_back(new UsbDaqDevice);
    }

    _location.siteNumber(s);
  }

  virtual ~MpsReadout() {
    for(unsigned i(0);i<_vUsbDaq.size();i++) delete _vUsbDaq[i];
  }

  virtual bool record(RcdRecord &r) {
    if(doPrint(r.recordType())) {
      std::cout << "MpsReadout::record()" << std::endl;
      r.RcdHeader::print(std::cout," ") << std::endl;
    }

    SubInserter inserter(r);

    // Check record type
    switch (r.recordType()) {

      // Run start
    case RcdHeader::runStart: {

      // Reset whole system and then readback run data
      for(unsigned j(0);j<_vUsbDaq.size();j++) assert(_vUsbDaq[j]->runReset());

      readRunData(r);

      // Setup configuration check
      for(unsigned i(0);i<3;i++) {
	for(unsigned j(0);j<_vUsbDaq.size();j++) {
	  _vConfigured[i].push_back(false);
	}
      }

      break;
    }
 
      // Configuration start is used to set up system
    case RcdHeader::configurationStart: {
      SubAccessor accessor(r);

      std::vector<const MpsReadoutConfigurationData*>
        c(accessor.access<MpsReadoutConfigurationData>());
      
      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  configurationStart"
		  << ", number of MpsReadoutConfigurationData objects found = "
		  << c.size() << std::endl << std::endl;
      }

      if(doPrint(r.recordType(),2)) {
	for(unsigned i(0);i<c.size();i++) {	  
	  std::cout << " MpsReadout::record()  configurationStart"
		    << ", MpsReadoutConfigurationData object "
		    << i << std::endl;
	  c[i]->print(std::cout,"  ") << std::endl;
	}
      }

      assert(c.size()==1);


      std::vector<const MpsLocationData<MpsUsbDaqConfigurationData>*>
        u(accessor.access< MpsLocationData<MpsUsbDaqConfigurationData> >());
      
      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  configurationStart"
		  << ", number of MpsUsbDaqConfigurationData objects found = "
		  << u.size() << std::endl << std::endl;
      }

      for(unsigned i(0);i<u.size();i++) {
	if(u[i]->write()) {
	  if(c[0]->usbDaqEnable((unsigned)u[i]->usbDaqNumber())) {
	    if(u[i]->usbDaqNumber()<_vUsbDaq.size()) {

	      if(doPrint(r.recordType(),2)) {
		std::cout << " MpsReadout::record()  configurationStart"
			<< ", writing MpsUsbDaqConfigurationData object "
			  << i << std::endl;
		u[i]->print(std::cout,"  ") << std::endl;
	      }
	      
	      assert(_vUsbDaq[u[i]->usbDaqNumber()]->
		     writeMpsUsbDaqConfigurationData(*(u[i]->data())));
	      _vConfigured[0][u[i]->usbDaqNumber()]=true;
	    }
	  }
	}
      }


      std::vector<const MpsLocationData<MpsPcb1ConfigurationData>*>
        w(accessor.access< MpsLocationData<MpsPcb1ConfigurationData> >());
      
      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  configurationStart"
		  << ", number of MpsPcb1ConfigurationData objects found = "
		  << w.size() << std::endl << std::endl;
      }

      for(unsigned i(0);i<w.size();i++) {
	if(w[i]->write()) {
	  if(c[0]->usbDaqEnable((unsigned)w[i]->usbDaqNumber())) {
	    if(w[i]->usbDaqNumber()<_vUsbDaq.size()) {
	      if(doPrint(r.recordType(),2)) {
		std::cout << " MpsReadout::record()  configurationStart"
			  << ", writing MpsPcb1ConfigurationData object "
			  << i << std::endl;
		w[i]->print(std::cout,"  ") << std::endl;
	      }
	      
	      assert(_vUsbDaq[w[i]->usbDaqNumber()]->
		     writeMpsPcb1ConfigurationData(*(w[i]->data())));
	      _vConfigured[1][w[i]->usbDaqNumber()]=true;
	    }
	  }
	}
      }


      std::vector<const MpsLocationData<MpsSensor1ConfigurationData>*>
        v(accessor.access< MpsLocationData<MpsSensor1ConfigurationData> >());
      
      MpsSensor1ConfigurationData dummy;
      
      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  configurationStart"
		  << ", number of MpsSensor1ConfigurationData objects found = "
		  << v.size() << std::endl << std::endl;
      }
      
      unsigned n(0);
      for(unsigned i(0);i<v.size();i++) {
	if(v[i]->write()) {
	  if(c[0]->usbDaqEnable((unsigned)v[i]->usbDaqNumber())) {
	    if(v[i]->usbDaqNumber()<_vUsbDaq.size()) {
	      if(doPrint(r.recordType(),2)) {
		std::cout << " MpsReadout::record()  configurationStart"
			  << ", writing MpsSensor1ConfigurationData object "
			  << i << std::endl;
		v[i]->print(std::cout,"  ") << std::endl;
	      }
	      
	      if(!_vConfigured[2][v[i]->usbDaqNumber()]) {
		assert(_vUsbDaq[v[i]->usbDaqNumber()]->loadMpsSensor1ConfigurationData(*(v[i]->data()),dummy));
		_vConfigured[2][v[i]->usbDaqNumber()]=true;
		
	      } else {
		MpsLocationData<MpsSensor1ConfigurationData>
		  *m(inserter.insert< MpsLocationData<MpsSensor1ConfigurationData> >());
		m->location(_location);
		m->usbDaqNumber(v[i]->usbDaqNumber());
		
		assert(_vUsbDaq[v[i]->usbDaqNumber()]->loadMpsSensor1ConfigurationData(*(v[i]->data()),*(m->data())));
		
		if(doPrint(r.recordType(),2)) {
		  std::cout << " MpsReadout::record()  configurationStart"
			    << ", read MpsSensor1ConfigurationData object "
			    << n << std::endl << std::endl;
		  m->print(std::cout,"  ") << std::endl;
		}
		n++;
	      }
	    }
	  }
	}
      }

      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  configurationStart"
		  << ", number of MpsSensor1ConfigurationData objects read = "
		  << n << std::endl << std::endl;
      }
      

      // Check all components have been configured
      for(unsigned j(0);j<_vUsbDaq.size();j++) {
	if(c[0]->usbDaqEnable(j)) {
	  assert(_vConfigured[0][j] && _vConfigured[1][j] && _vConfigured[2][j]);
	}
      }
      
      // Now read back the (non-destructive read) configuration data
      readConfigurationData(r);

      break;
    }
      
    case RcdHeader::bunchTrain: {

      // USB_DAQ number 0 is always master
      assert(_vUsbDaq[0]->takeBunchTrain());

      // Read out USB_DAQ data
      unsigned n(0);
      for(unsigned i(0);i<_vUsbDaq.size();i++) {
	MpsLocationData<MpsUsbDaqBunchTrainData>
	  *m(inserter.insert< MpsLocationData<MpsUsbDaqBunchTrainData> >());
	m->location(_location);
	m->usbDaqNumber(i);
	
	assert(_vUsbDaq[i]->readMpsUsbDaqBunchTrainData(*(m->data())));
	inserter.extend(4*m->data()->numberOfTrgs());

	if(doPrint(r.recordType(),2)) {
	  std::cout << " MpsReadout::record()  bunchTrain"
		    << ", read MpsUsbDaqBunchTrainData object "
		    << n << std::endl << std::endl;
	  m->print(std::cout,"  ") << std::endl;
	}
	n++;
      }

      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  bunchTrain"
		  << ", number of MpsUsbDaqBunchTrainData objects read = "
		  << n << std::endl << std::endl;
      }


      // Read out sensor PCB data
      n=0;
      for(unsigned i(0);i<_vUsbDaq.size();i++) {
	 MpsLocationData<MpsPcb1BunchTrainData>
	  *m(inserter.insert< MpsLocationData<MpsPcb1BunchTrainData> >());
	m->location(_location);
	m->usbDaqNumber(i);

	assert(_vUsbDaq[i]->readMpsPcb1BunchTrainData(*(m->data())));

	if(doPrint(r.recordType(),2)) {
	  std::cout << " MpsReadout::record()  bunchTrain"
		    << ", read MpsPcb1BunchTrainData object "
		    << n << std::endl << std::endl;
	  m->print(std::cout,"  ") << std::endl;
	}
	n++;
      }

      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  bunchTrain"
		  << ", number of MpsPcb1BunchTrainData objects read = "
		  << n << std::endl << std::endl;
      }


      // Read out sensor data
      n=0;
      for(unsigned i(0);i<_vUsbDaq.size();i++) {
	 MpsLocationData<MpsSensor1BunchTrainData>
	  *m(inserter.insert< MpsLocationData<MpsSensor1BunchTrainData> >());
	m->location(_location);
	m->usbDaqNumber(i);

	assert(_vUsbDaq[i]->readMpsSensor1BunchTrainData(*(m->data())));
	inserter.extend(4*m->data()->totalNumberOfHits());

	if(doPrint(r.recordType(),2)) {
	  std::cout << " MpsReadout::record()  bunchTrain"
		    << ", read MpsSensor1BunchTrainData object "
		    << n << std::endl << std::endl;
	  m->print(std::cout,"  ") << std::endl;
	}
	n++;
      }
      
      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  bunchTrain"
		  << ", number of MpsSensor1BunchTrainData objects read = "
		  << n << std::endl << std::endl;
      }

      break;
    }

    case RcdHeader::configurationEnd: {
      readConfigurationData(r);
      break;
    }
 
      // Run end
    case RcdHeader::runEnd: {
      readRunData(r);

      MpsSensor1ConfigurationData dummy;

      unsigned n(0);
      for(unsigned i(0);i<_vUsbDaq.size();i++) {
	MpsLocationData<MpsSensor1ConfigurationData>
	  *m(inserter.insert< MpsLocationData<MpsSensor1ConfigurationData> >());
	m->location(_location);
	m->usbDaqNumber(i);

	assert(_vUsbDaq[i]->loadMpsSensor1ConfigurationData(dummy,*(m->data())));
	if(doPrint(r.recordType(),2)) {
	  std::cout << " MpsReadout::record()  runEnd"
		    << ", read MpsSensor1ConfigurationData object "
		    << n << std::endl << std::endl;
	  m->print(std::cout,"  ") << std::endl;
	}
	n++;
      }

      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  runEnd"
		  << ", number of MpsSensor1ConfigurationData objects read = "
		  << n << std::endl << std::endl;
      }

      break;
    }
 
      // Slow readout
    case RcdHeader::slowReadout: {

      // Read out sensor PCB data as only temperatures
      unsigned n(0);
      for(unsigned i(0);i<_vUsbDaq.size();i++) {
	 MpsLocationData<MpsPcb1BunchTrainData>
	  *m(inserter.insert< MpsLocationData<MpsPcb1BunchTrainData> >());
	m->location(_location);
	m->usbDaqNumber(i);

	assert(_vUsbDaq[i]->readMpsPcb1BunchTrainData(*(m->data())));

	if(doPrint(r.recordType(),2)) {
	  std::cout << " MpsReadout::record()  slowReadout"
		    << ", read MpsPcb1BunchTrainData object "
		    << n << std::endl << std::endl;
	  m->print(std::cout,"  ") << std::endl;
	}
	n++;
      }

      if(doPrint(r.recordType(),1)) {
	std::cout << " MpsReadout::record()  slowReadout"
		  << ", number of MpsPcb1BunchTrainData objects read = "
		  << n << std::endl << std::endl;
      }

      break;
    }

    default: {
      break;
    }
    };
    
    return true;
  }

private:
  bool readRunData(RcdRecord &r) {
    SubInserter inserter(r);

    unsigned n(0);
    for(unsigned i(0);i<_vUsbDaq.size();i++) {
      MpsLocationData<MpsUsbDaqRunData>
	*m(inserter.insert< MpsLocationData<MpsUsbDaqRunData> >());
      m->location(_location);
      m->usbDaqNumber(i);

      assert(_vUsbDaq[i]->readMpsUsbDaqRunData(*(m->data())));

      if(doPrint(r.recordType(),2)) {
	std::cout << " MpsReadout::readRunData()"
		  << "  Read MpsUsbDaqRunData object " << n << std::endl;
	m->print(std::cout," ") << std::endl;
      }

      n++;
    }

    if(doPrint(r.recordType(),1)) {
      std::cout << " MpsReadout::readRunData()"
		<< "  Number of MpsUsbDaqRunData objects read = "
		<< n << std::endl << std::endl;
    }

    return true;
  }

  bool readConfigurationData(RcdRecord &r) {
    SubInserter inserter(r);

    // Read back all the USB_DAQ configuration data

    unsigned n(0);
    for(unsigned i(0);i<_vUsbDaq.size();i++) {
      MpsLocationData<MpsUsbDaqConfigurationData>
	*m(inserter.insert< MpsLocationData<MpsUsbDaqConfigurationData> >());
      m->location(_location);
      m->usbDaqNumber(i);

      assert(_vUsbDaq[i]->readMpsUsbDaqConfigurationData(*(m->data())));

      if(doPrint(r.recordType(),2)) {
	std::cout << " MpsReadout::readConfigurationData()"
		  << "  Read MpsUsbDaqConfigurationData object " << n << std::endl;
	m->print(std::cout," ") << std::endl;
      }

      n++;
    }

    if(doPrint(r.recordType(),1)) {
      std::cout << " MpsReadout::readConfigurationData()"
		<< "  Number of MpsUsbDaqConfigurationData objects read = "
		<< n << std::endl << std::endl;
    }

    // Repeat for all the PCB1 configuration data

    n=0;
    for(unsigned i(0);i<_vUsbDaq.size();i++) {
      MpsLocationData<MpsPcb1ConfigurationData>
	*m(inserter.insert< MpsLocationData<MpsPcb1ConfigurationData> >());
      m->location(_location);
      m->usbDaqNumber(i);

      assert(_vUsbDaq[i]->readMpsPcb1ConfigurationData(*(m->data())));

      if(doPrint(r.recordType(),2)) {
	std::cout << " MpsReadout::readConfigurationData()"
		  << "  Read MpsPcb1ConfigurationData object " << n << std::endl;
	m->print(std::cout," ") << std::endl;
      }

      n++;
    }

    if(doPrint(r.recordType(),1)) {
      std::cout << " MpsReadout::readConfigurationData()"
		<< "  Number of MpsPcb1ConfigurationData objects read = "
		<< n << std::endl << std::endl;
    }

    return true;
  }



  MpsLocation _location;
  std::vector<UsbDaqDevice*> _vUsbDaq;
  std::vector<bool> _vConfigured[3];
};

#endif
